from typing import Union, List, Dict, Tuple
from privex.jsonrpc.core import atomic_to_decimal, decimal_to_atomic
from decimal import Decimal


class ObjBase:
    """
    A base class to be extended by data storage classes, allowing their attributes to be
    accessed as if the class was a dict/list.
    
    Also allows the class to be converted into a dict/list if raw_data is filled, like so: ``dict(SomeClass())``
    """

    def __init__(self, raw_data: Union[list, tuple, dict] = None, *args, **kwargs):
        self.raw_data = {} if not raw_data else raw_data  # type: Union[list, tuple, dict]
        super(ObjBase, self).__init__(raw_data, *args, **kwargs)

    def __iter__(self):
        r = self.raw_data
        if type(r) is dict:
            for k, v in r.items(): yield (k, v,)
            return
        for k, v in enumerate(r): yield (k, v,)

    def __getitem__(self, key):
        """
        When the instance is accessed like a dict, try returning the matching attribute.
        If the attribute doesn't exist, or the key is an integer, try and pull it from raw_data
        """
        if type(key) is int: return self.raw_data[key]
        if hasattr(self, key): return getattr(self, key)
        if key in self.raw_data: return self.raw_data[key]
        raise KeyError(key)

    @classmethod
    def from_list(cls, obj_list: List[dict]):
        """
        Converts a ``list`` of ``dict`` 's into a ``Generator[cls]`` of instances of the class you're calling this from.

        **Example:**

            >>> _balances = [dict(account='someguy123', symbol='SGTK', balance='1.234')]
            >>> balances = list(SEBalance.from_list(_balances))
            >>> type(balances[0])
            <class 'privex.steemengine.objects.SEBalance'>
            >>> balances[0].account
            'someguy123'
        
        """
        for tx in obj_list:
            yield cls(**tx)


class MoneroPayment(ObjBase):
    """
    Represents a Payment from ``get_payments`` in Monero
    """

    def __init__(self, address: str = "", amount: int = 0, block_height: int = None, payment_id="",
                 subaddr_index: dict = {}, tx_hash="", unlock_time: int = 0, **kwargs):
        self.address, self.amount, self.block_height, self.payment_id = address, amount, block_height, payment_id
        self.subaddr_index, self.tx_hash, self.unlock_time = subaddr_index, tx_hash, unlock_time
        self.raw_data = {
            **kwargs,
            **dict(
                address=address, amount=amount, block_height=block_height, payment_id=payment_id,
                subaddr_index=subaddr_index, tx_hash=tx_hash, unlock_time=unlock_time
            )
        }

    @property
    def decimal_amount(self) -> Decimal:
        """Return the atomic integer :py:attr:`.amount` as a normal Decimal amount"""
        return atomic_to_decimal(self.amount)


class MoneroTransfer(ObjBase):
    """
    Represents a Transfer item from ``get_transfers`` in Monero
    """

    def __init__(self, address: str = "", amount: int = 0, confirmations: int = 0, fee: int = 0,
                 height: int = 0, payment_id="", subaddr_index: dict = None, subaddr_indices: List[dict] = None,
                 timestamp: int = None, txid: str = "", type: str = "", unlock_time: int = 0, note: str = None,
                 suggested_confirmations_threshold: int = 0, double_spend_seen: bool = False, **kwargs):

        if subaddr_indices is None:
            subaddr_indices = []
        if subaddr_index is None:
            subaddr_index = {}

        self.address = address
        self.amount = amount
        self.confirmations = confirmations
        self.double_spend_seen = double_spend_seen
        self.fee = fee
        self.height = height
        self.note = note
        self.payment_id = payment_id
        self.subaddr_index = subaddr_index
        self.subaddr_indices = subaddr_indices
        self.timestamp = timestamp
        self.txid = txid
        self.type = type
        self.unlock_time = unlock_time
        self.suggested_confirmations_threshold = suggested_confirmations_threshold

        dicdata = dict(
            address=address, amount=amount, confirmations=confirmations, double_spend_seen=double_spend_seen, fee=fee,
            height=height, note=note, payment_id=payment_id, subaddr_index=subaddr_index, unlock_time=unlock_time,
            subaddr_indices=subaddr_indices, timestamp=timestamp, txid=txid, type=type,
            suggested_confirmations_threshold=suggested_confirmations_threshold
        )

        self.raw_data = {**kwargs, **dicdata}

    @property
    def decimal_amount(self) -> Decimal:
        """Return the atomic integer :py:attr:`.amount` as a normal Decimal amount"""
        return atomic_to_decimal(self.amount)

    @property
    def decimal_fee(self) -> Decimal:
        """Return the atomic integer :py:attr:`.fee` as a normal Decimal amount"""
        return atomic_to_decimal(self.fee)

    def __str__(self):
        return f'<MoneroTransfer amount="{self.decimal_amount} XMR", address="{self.address}" ' \
               f'confirmations={self.confirmations} >'

    def __repr__(self):
        return self.__str__()




from privex.helpers import empty

from privex.jsonrpc.JsonRPC import JsonRPC
from decimal import Decimal
from typing import Union, List, Dict
from privex.jsonrpc.objects import MoneroPayment, MoneroTransfer
from privex.jsonrpc.core import decimal_to_atomic, atomic_to_decimal


class MoneroRPC(JsonRPC):
    """
    Monero JsonRPC Client for Python

    Basic Usage:

        >>> m = MoneroRPC(username='monero', password='mypass')
        >>> m.open_wallet(filename='mywallet', password='mywalletpass')
        {}
        >>> m.get_height()
        173463
        >>> m.create_account(label='myaccount')
        {"account_index": 1, "address": "77Vx9cs1VPicFndSVgYUvTdLCJEZw9h81hXL..."}
        >>> for p in m.get_payments('60900e5603bf96e3'):
        ...     print(p.address, ':', p.decimal_amount)
        55LTR8KniP4LQGJSPtbYDacR7dz8RBFnsfAKMa : 1.2345
        55LTR8KniP4LQGJSPtbYDacR7dz8RBFnsfAKMa : 2.342
        >>> m.make_uri(address='77Vx9cs1VPicFndSVgYUvTdLCJEZw9h81hXL...', amount=m.decimal_to_atomic('10'))
        monero:77Vx9cs1VPicFndSVgYUvTdLCJEZw9h81hXL...?tx_amount=10000000000000


    Copyright::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Python Simple JSON RPC library             |
        |        License: X11/MIT                           |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |                                                   |
        +===================================================+
    
    """
    DEF_URL = '/json_rpc'

    def __init__(self, hostname='127.0.0.1', port=18082, username=None, password=None, ssl=False, timeout=120, 
                 url=DEF_URL, auth='digest'):
        super().__init__(
            hostname=hostname, port=port, username=username, password=password, 
            ssl=ssl, timeout=timeout, url=url, auth=auth
        )
    
    @staticmethod
    def decimal_to_atomic(amount: Union[Decimal, str]) -> int:
        """
        Convert a Decimal monero amount into an atomic (piconero) integer amount

            >>> MoneroRPC.decimal_to_atomic(Decimal('10'))
            10000000000000
        
        """
        return decimal_to_atomic(amount)
    
    @staticmethod
    def atomic_to_decimal(amount: int) -> Decimal:
        """
        Convert an integer atomic monero amount into a normal Decimal amount
        
            >>> MoneroRPC.atomic_to_decimal(10000000000000)
            Decimal('10')
        
        """
        return atomic_to_decimal(amount)
    
    def create_address(self, account_index: int = 0, label: str = "") -> dict:
        """
        Create a monero address. Returns a dictionary with the monero address ``address`` and ID ``address_index``

        :param int account_index: The account ID to create an address for
        :param str label: An optional label for the address
        :return dict addr: dict(address: str, address_index: int)
        """
        return self.call('create_address', account_index=account_index, label=label)

    def create_account(self, label: str = "") -> dict:
        """
        Creates an account with the given optional label, and returns the first address + account index as a dict.

        :return dict acc: dict(address: str, account_index: 2)
        """
        return self.call('create_account', label=label)
    
    def create_wallet(self, filename: str, password: str = None, language="English") -> dict:
        if password is None:
            return self.call('create_wallet', filename=filename, language=language)
        return self.call('create_wallet', filename=filename, password=password, language=language)

    def get_height(self) -> int:
        """Returns the current block height of the monero wallet/rpc as an integer"""
        return self.call('get_height').get('height')
    
    def get_accounts(self, tag: str = None) -> dict:
        """
        Returns a dict containing a list of accounts, and total balances.

        Subaddress accounts are formatted as: ``dict(account_index: int, balance: int, base_address: str, label: str, tag: str, unlocked_balance: int)``
        
        :return dict accounts: dict(subaddress_accounts: List[dict], total_balance: int, total_unlocked_balance: int)
        """
        return self.call('get_accounts') if tag is None else self.call('get_accounts', tag=tag)
    
    def get_address(self, account_index=0, address_index: List[int] = []):
        return self.call('get_address', account_index=account_index, address_index=address_index)
    
    def get_balance(self, account_index=0, address_indices: List[int] = []) -> dict:
        """
        Get the balance for the wallet

        :return dict baldata: dict(balance: int, unlocked_balance: int, multisig_import_needed: bool, per_subaddress: List[dict])
        """
        return self.call('get_balance', account_index=account_index, address_indices=address_indices)

    def get_payments(self, payment_id: str) -> List[MoneroPayment]:
        p = self.call('get_payments', payment_id=payment_id)
        payments = p.get('payments', [])
        return list(MoneroPayment.from_list(payments))

    def get_transfers(self, account_index=0, pending=True, incoming=True, outgoing=True,
                      subaddr_indices: List[int] = None, **kw) -> Dict[str, List[MoneroTransfer]]:
        """
        Get incoming/outgoing transfers as a ``dict`` - generally with the keys ``in`` and ``out``

        See :class:`.MoneroTransfer` for the attributes contained in each transfer object.

        NOTE: As ``in`` is a protected Python keyword, ``in`` and ``out`` have been changed to ``incoming``
        and ``outgoing``, however you can still use ``in`` and ``out`` if they're passed as kwargs using
        ``**somedict`` with a quoted dictionary
        e.g. ``get_transfers(account_index=1, **{"in": False, "out": True})``

        Usage:

            >>> m = MoneroRPC(username='monero', password='somepass')
            >>> transfers = m.get_transfers(account_index=1)
            >>> for t in transfers['in']:   # type: MoneroTransfer
            ...     print(f'Received {t.decimal_amount:.4} XMR into address {t.address}')
            ...
            Received 1.000 XMR into address 77Vx9cs1VPicFndSVgYUvTdLCJEZw9h81hXL...
            Received 2.473 XMR into address 55LTR8KniP4LQGJSPtbYDacR7dz8RBFnsfAKMa...
            >>>


        :param int account_index: (Default: 0) The account ID to get transfers for
        :param bool pending:  (Default: True) Include unconfirmed transactions
        :param bool incoming: (Default: True) List incoming transfers (``in``)
        :param bool outgoing: (Default: True) List outgoing transfers (``out``)
        :param list subaddr_indices: (Optional) A list of int sub-address indices to filter transfers for
        :param Any kw: Any additional dictionary parameters to pass to get_transfers
        :return dict transfers: dict(in: List[MoneroTransfer], out: List[MoneroTransfer])
        """
        _args = {
            "account_index": account_index, "pending": pending,
            "in": kw.get('in', incoming), "out": kw.get('out', outgoing),
            "subaddr_indices": [] if empty(subaddr_indices, itr=True) else subaddr_indices, **kw
        }
        args = {k: v for k, v in _args.items() if v is not None}
        t = self.call('get_transfers', **args)
        res = {}
        if 'in' in t:
            res['in'] = list(MoneroTransfer.from_list(t['in']))
        if 'out' in t:
            res['out'] = list(MoneroTransfer.from_list(t['out']))
        return {**res, **{k: v for k, v in t.items() if k not in ['in', 'out']}}

    def get_version(self) -> int:
        """
        Returns the RPC version as an integer, formatted with Major * 2^16 + Minor (Major encoded over the first 16 bits, 
        and Minor over the last 16 bits).
        """
        return self.call('get_version').get('version')
    
    def make_uri(self, address, amount: int = None, payment_id: str = None, recipient_name=None, tx_description=None) -> str:
        """Returns a Monero payment URI based on the arguments specified"""
        _args = dict(address=address,amount=amount,payment_id=payment_id,recipient_name=recipient_name,tx_description=tx_description)
        args = {k: v for k, v in _args.items() if v is not None}
        return self.call('make_uri', **args).get('uri', '')
    
    def make_integrated_address(self, standard_address: str = None, payment_id: str = None) -> dict:
        """
        Creates an integrated address (monero address combined with payment id)

        :param standard_address: A monero address (if not specified, will default to the "primary address")
        :param payment_id: A payment ID to integrate (if not specified, a random one will be generated)
        :return dict result: dict(integrated_address: str, payment_id: str)
        """
        _args = dict(standard_address=standard_address, payment_id=payment_id)
        args = {k: v for k, v in _args.items() if v is not None}
        return self.call('make_integrated_address', **args)

    def open_wallet(self, filename: str, password: str = None) -> dict:
        """Open a Monero wallet by filename. Optionally specify a password if it's required"""
        if password is None:
            return self.call('open_wallet', filename=filename)
        return self.call('open_wallet', filename=filename, password=password)

    def close_wallet(self) -> dict:
        """Close the current wallet (returns an empty dict)"""
        return self.call('close_wallet')

    def store(self) -> dict:
        """Save the wallet file"""
        return self.call('store')

    def transfer(self, destinations: List[dict], account_index: int = 0, subaddr_indices: List[int] = [], priority: int = 0, 
                 mixin: int = 0, ring_size: int = 0, payment_id: str = "", **kwargs) -> dict:
        """
        Transfer monero to one or more destinations.

        Basic usage (monero address and tx_key truncated for sanity):

            >>> m = MoneroRPC(username='monero', password='somepass')
            >>> amt = m.decimal_to_atomic('10')
            >>> m.transfer(destinations=[dict(amount=amt, address='7BnERTpvL5MbCLtj5n9No7J5oE5hH...')])
            {
                "amount": 10000000000000,
                "fee": 86897600000,
                "multisig_txset": "",
                "tx_blob": "",
                "tx_hash": "7663438de4f72b25a0e395b770ea9ecf7108cd2f0c4b75be0b14a103d3362be9",
                "tx_key": "25c9d8ec20045c80c93d665c9d3684aab7335f8b2cd02e1ba263...",
                "tx_metadata": "",
                "unsigned_txset": ""
            }
        

        """
        return self.call(
            'transfer', 
            destinations=destinations, account_index=account_index, subaddr_indices=subaddr_indices,
            priority=priority, mixin=mixin, ring_size=ring_size, payment_id=payment_id, **kwargs
        )
    
    def simple_transfer(self, amount: Union[Decimal, str], address: str, account_index: int = 0, payment_id: str = "", **kwargs):
        """
        A small wrapper function for :py:meth:`.transfer` to make sending Monero easier.

        Usage (monero address and tx_key truncated for sanity):

            >>> m = MoneroRPC(username='monero', password='somepass')
            >>> m.simple_transfer('10', '7BnERTpvL5MbCLtj5n9No7J5oE5hH...')
            {
                "amount": 10000000000000,
                "fee": 86897600000,
                "multisig_txset": "",
                "tx_blob": "",
                "tx_hash": "7663438de4f72b25a0e395b770ea9ecf7108cd2f0c4b75be0b14a103d3362be9",
                "tx_key": "25c9d8ec20045c80c93d665c9d3684aab7335f8b2cd02e1ba263...",
                "tx_metadata": "",
                "unsigned_txset": ""
            }
        """
        atomic_amt = self.decimal_to_atomic(amount)
        dest = [dict(amount=atomic_amt, address=address)]
        return self.transfer(destinations=dest, account_index=account_index, payment_id=payment_id, **kwargs)

    def validate_address(self, address: str) -> dict:
        """
        Validate a monero address

            >>> m = MoneroRPC(username='monero', password='somepass')
            >>> m.validate_address(
            ...    '8BNkere1xg7Jsbth2HbGBMdzFvgLDxpaJ4nmqqk72GFwVKGWpgPv3Xib1bEEx8hMt1PtpF3BiZws4dEovurWD73NDh9gFDj'
            ... )
            {'integrated': False, 'nettype': 'mainnet', 'openalias_address': '', 'subaddress': True, 'valid': True}


        :param str address: A monero address
        :return dict addr_info: dict(integrated:bool, nettype:str, openalias_address:str, subaddress:bool, valid:bool)
        """
        return self.call('validate_address', address=address)




from decimal import Decimal
from typing import Union

ATOMIC_UNIT = int(1e12)

def decimal_to_atomic(amount: Union[Decimal, str]) -> int:
    """
    Convert a Decimal monero amount into an atomic (piconero) integer amount

        >>> decimal_to_atomic(Decimal('10'))
        10000000000000
    
    """
    return int(Decimal(amount) * Decimal(ATOMIC_UNIT))

def atomic_to_decimal(amount: int) -> Decimal:
    """
    Convert an integer atomic monero amount into a normal Decimal amount
    
        >>> atomic_to_decimal(10000000000000)
        Decimal('10')
    
    """
    return Decimal(amount) / Decimal(ATOMIC_UNIT)




import json
import requests
import logging
from requests.auth import HTTPDigestAuth
from json import JSONDecodeError
from decimal import Decimal
from typing import Union, List, Dict

log = logging.getLogger(__name__)


class RPCException(BaseException):
    """Thrown when 'error' is present in the result, and is not None/False"""
    pass


class JsonRPC:
    """
    JsonRPC - a small Python class for querying JSON RPC servers over HTTP / HTTPS

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Python Simple JSON RPC library             |
        |        License: X11/MIT                           |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |                                                   |
        +===================================================+
    
    Basic usage:

        >>>  j = JsonRPC(hostname='api.example.com', port=443, ssl=True)
        >>>  j.get_someinfo('my first parameter')
        ['some', 'result', 'data', 'returned']
    
    If a method is not defined, you can still use it! You just won't get any IDE hints with the parameters.
    """

    LAST_ID = 0

    # Using a python-requests session will help to encourage connection re-use, e.g. HTTP Keep-alive
    # Since this is a static attribute, it will be shared across ALL instances of the class.
    req = requests.Session()

    def __init__(self, hostname, port: int = 80, username=None, password=None, ssl=False, timeout=120, 
                 url: str = '', auth: str = 'plain'):
        """
        Configure the remote JSON RPC server settings

        :param hostname: The hostname or IP address of the JSON RPC server
        :param port:     The JSON RPC TCP port to connect to
        :param username: If the RPC server needs a username, specify it here
        :param password: If the RPC server needs a password, specify it here (username must also be set)
        :param ssl:      If set to True, will use https for requests. Default is false - use plain http
        :param timeout:  If the server stops sending us data for this many seconds, abort and throw an exception
        :param url:      The URL to query, e.g. api/v1/test (starting /'s will automatically be removed)
        :param str auth: HTTP Authentication type - either ``plain`` (default) or ``digest``
        """
        self.timeout = timeout
        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self.ssl = ssl
        self.endpoint = url
        self.auth = auth = auth.lower()
        if auth not in ['plain', 'digest']:
            raise AttributeError('Attribute "auth" must be either "plain" or "basic".')
        self.headers = {'content-type': 'application/json'}
        self.method_base = ''

    @property
    def url(self):
        url = self.endpoint
        proto = 'https' if self.ssl else 'http'
        host = '{}:{}'.format(self.hostname, self.port)
        if self.username is not None and self.auth == 'plain':
            host = '{}:{}@{}:{}'.format(self.username, self.password, self.hostname, self.port)
        url = url[1:] if len(url) > 0 and url[0] == '/' else url  # Strip starting / of URL

        return "{}://{}/{}".format(proto, host, url)

    @property
    def next_id(self):
        JsonRPC.LAST_ID += 1
        return JsonRPC.LAST_ID

    def add_headers(self, custom_headers):
        self.headers.update(custom_headers)

    def set_method_base(self, method_base):
        self.method_base = method_base
    
    def _call(self, method: str, params: Union[dict, list] = [], jid: int = None, del_params=False, 
              raise_status=True) -> Union[dict, list]:
        """
        This is an internal function used by :py:meth:`.call` which handles the actual JsonRPC call.

        Some of the few reasons that you may wish to use ``_call`` directly instead of :py:meth:`.call` :
        
         - Need to specify parameters directly, e.g. blank JSON dict params ``{}``
         - Need to control the JsonRPC request ID
         - Need to read the original JsonRPC JSON response, not just the ``result`` key
         - Need to ignore or have custom handling for the ``error`` field in the response
        
        Examples:

            >>> j = JsonRPC('127.0.0.1', 8080)
            >>> # Simple call
            >>> res = j._call('get_user', ['john'])
            >>> print(res['result'])
            {'name': 'john', 'age': 30}
            >>> # Call with dict params, and custom JsonRPC request ID
            >>> res = j._call('my_query', dict(user='dave', age=25), jid=3)
            >>> print(res['id'])
            3

        :param str method:               JSON RPC method to call
        :param Union[dict,list] params:  Parameters to be passed as 'params' in the JSON request body (will not be casted)
        :param int jid:                  Optional JSON Request ID, if not specified will use :py:meth:`.next_id`
        :param bool del_params:          (Default: ``False``) If True, will remove ``params`` entirely from the request.
        :param bool raise_status:        (Default: ``True``) If True, will call ``.raise_for_status()`` after the request, 
                                         causing a HTTPError to be raised for non-200 responses
        
        :raises requests.exceptions.HTTPError: If ``raise_status`` is True, HTTPError will be raised for non-200 responses.
        
        :return Union[dict,list] result: The parsed JSON response as a dict/list (_call requires you to extract ``result`` yourself)
        """
        headers = self.headers

        payload = {
            "method": method,
            "params": params,
            "jsonrpc": "2.0",
            "id": self.next_id if jid is None else jid,
        }
        if del_params:   # If del_params is True, then we need to remove 'params' from the request completely.
            del payload['params']

        r = None
        try:
            log.debug('Sending JsonRPC request to %s with payload: %s', self.url, payload)
            req_args = dict(data=json.dumps(payload), headers=headers, timeout=self.timeout)
            if self.auth == 'digest' and self.username is not None:
                req_args['auth'] = HTTPDigestAuth(self.username, self.password)
            r = self.req.post(self.url, **req_args)
            if raise_status:
                r.raise_for_status()
            response = r.json()
            return response
        except JSONDecodeError as e:
            log.warning('JSONDecodeError while querying %s', self.url)
            log.warning('Params: %s', params)
            t = r.text.decode('utf-8') if type(r.text) is bytes else str(r.text)
            log.warning('Raw response data was: %s', t)
            raise e

    def call(self, method, *params, **dicdata) -> Union[dict, list]:
        """
        Calls a JSON RPC method with method 'params' being the positional args passed as a list.

        If keyword args are passed, params will be a dict of the kwargs `dicdata` and positional args will be ignored.

        If both positional and keyword args are empty, params will be set to an empty list []

        :param method: JSON RPC method to call
        :param params: Parameters to be passed as a list as 'params' in the JSON request body
        :param dicdata: Arguments to be passed as a dict as 'params' in the JSON request body (overrides params)

        :raises RPCException: When an RPC call returns with a non-null/non-false 'error' key
        :raises requests.exceptions.HTTPError: Raised if the JsonRPC call resulted in a non-200 response.

        :return: dict() or list() of results, depending on what format the method returns.
        """
        
        method = self.method_base+'/'+method if self.method_base else method
        
        # If kwargs are passed, payload params is a dictionary of the kwargs instead of positionals
        _params = dict(dicdata) if len(dicdata.keys()) > 0 else list(params)
        response = self._call(method=method, params=_params, raise_status=True)

        if 'error' in response and response['error'] not in [None, False]:
            raise RPCException(response['error'])
        return response['result']

    def __getattr__(self, name):
        """
        Methods that haven't yet been defined are simply passed off to :meth:`.call` with the positional and kwargs.

        This means `rpc.getreceivedbyaddress(address)` is equivalent to `rpc.call('getreceivedbyaddress', address)`

        :param name: Name of the attribute requested
        :return: Dict or List from call result
        """
        def c(*args, **kwargs):
            return self.call(name, *args, **kwargs)
        return c

"""
+===================================================+
|                 © 2019 Privex Inc.                |
|               https://www.privex.io               |
+===================================================+
|                                                   |
|        Python Simple JSON RPC library             |
|        License: X11/MIT                           |
|                                                   |
|        Core Developer(s):                         |
|                                                   |
|          (+)  Chris (@someguy123) [Privex]        |
|                                                   |
+===================================================+

Python Json RPC - A simple library for interacting with JsonRPC services
Copyright (c) 2019    Privex Inc. ( https://www.privex.io )

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation 
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, 
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the 
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE 
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS 
OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name(s) of the above copyright holders shall not be used in advertising or 
otherwise to promote the sale, use or other dealings in this Software without prior written authorization.
"""
